#version 430 core
/*
* Copyright 2018 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/
#define MAX_DirectionalLights 32
#define MAX_PointLights 32
#define MAX_SpotLights 32

in vec3 position_world;
in vec3 newNormal;
in vec2 TexCoord;


uniform vec3 cameraPos;

uniform int numberOfDLights;
uniform int numberOfPLights;
uniform int numberOfSLights;

uniform float objectAlpha = 1;

struct Material {
	sampler2D texture_diffuse0;
	sampler2D texture_diffuse1;
	sampler2D texture_diffuse2;
	sampler2D texture_specular0;
	sampler2D texture_specular1;
	vec3 diffuseColor;
	float k_Ambient;
	float k_Diffuse;
	float k_Specular;
	float alpha;
};

struct AmbientLight {
	vec3 intensity;
};

struct DirectionalLight {
	vec3 intensity;
	vec3 direction;
	
	bool shadowMapAcitve;
	sampler2D shadowMap;
	mat4 lightSpaceMatrix;
};

struct PointLight {
	vec3 intensity;
	vec3 position;
	float constant;
	float linear;
	float quadratic;
};

struct SpotLight {
	vec3 intensity;
	vec3 position;
	vec3 direction;
	float innerAngle;
	float outerAngle;
	float constant;
	float linear;
	float quadratic;
};

uniform Material material;
uniform AmbientLight ambientLight;
uniform DirectionalLight directionalLight[MAX_DirectionalLights];
uniform PointLight pointLight[MAX_PointLights];
uniform SpotLight spotLight[MAX_SpotLights];
uniform float brightnessMultiplier = 1.0f;
uniform bool doSpecularMap;

out vec4 color;


// function declarations
vec3 calcDirLight(DirectionalLight directionalLight[MAX_DirectionalLights], vec3 normalizedNormal, vec3 view, float specularMapFactor);
vec3 calcPointLight(PointLight pointLight[MAX_PointLights], vec3 normalizedNormal, vec3 view);
vec3 calcSpotLight(SpotLight spotLight[MAX_SpotLights], vec3 normalizedNormal, vec3 view);


void main() {	
	vec3 normalizedNormal = normalize(newNormal);
	vec3 view = normalize(cameraPos - position_world);

	float specularMapFactor = texture(material.texture_specular0, TexCoord).r;
	if(specularMapFactor == 0 || !doSpecularMap) specularMapFactor = material.k_Specular;
	vec3 ambient = texture(material.texture_diffuse0, TexCoord).rgb * ambientLight.intensity * material.k_Ambient;
	vec3 dLight = calcDirLight(directionalLight, normalizedNormal, view, specularMapFactor);
	vec3 pLight = calcPointLight(pointLight, normalizedNormal, view);
	vec3 sLight = calcSpotLight(spotLight, normalizedNormal, view);

	color = vec4((ambient + dLight + pLight + sLight) * brightnessMultiplier, texture(material.texture_diffuse0, TexCoord).a * objectAlpha);// - vec4(texture(material.texture_diffuse1, TexCoord).rrr,0);
	//color = vec4(view, 1);
}


vec3 calcDirLight(DirectionalLight directionalLight[MAX_DirectionalLights], vec3 normalizedNormal, vec3 view, float specularMapFactor) {
	vec3 dLight = vec3(0.0f, 0.0f, 0.0f);

	for(int i = 0; i < numberOfDLights; i++) {
		if(directionalLight[i].shadowMapAcitve) {
			vec4 toLightSpace = directionalLight[i].lightSpaceMatrix * vec4(position_world, 1.0f);
			vec3 posInLightSpace = toLightSpace.xyz / toLightSpace.w;
			posInLightSpace = posInLightSpace * 0.5 + 0.5;

			float dist = posInLightSpace.z;
			float lightDist = texture(directionalLight[i].shadowMap, posInLightSpace.xy).r;
			//if(dist > lightDist) continue;		//this activates shadows
		}
		vec3 r = normalize(reflect(directionalLight[i].direction, normalizedNormal));
		vec3 d_diffuse = directionalLight[i].intensity * material.k_Diffuse * max(0.0f, dot(-normalize(directionalLight[i].direction), normalizedNormal));
		vec3 d_specular = directionalLight[i].intensity * specularMapFactor * pow(max(0.0f, dot(r, normalize(view))), material.alpha);
		dLight += (texture(material.texture_diffuse0, TexCoord).rgb * d_diffuse + d_specular);
	}
	return dLight;
}


vec3 calcPointLight(PointLight pointLight[MAX_PointLights], vec3 normalizedNormal, vec3 view) {
	vec3 pLight = vec3(0.0f, 0.0f, 0.0f);

	for(int i = 0; i < numberOfPLights; i++){
		vec3 pointLightDirection = position_world - pointLight[i].position;
		vec3 r2 = normalize(reflect(pointLightDirection, normalizedNormal));
		float d = distance(position_world, pointLight[i].position);
		float attenuation = pointLight[i].constant + d * pointLight[i].linear + pow(d, 2) * pointLight[i].quadratic;
		vec3 p_diffuse = pointLight[i].intensity * (1/attenuation) * material.k_Diffuse * max(0.0f, dot(-normalize(pointLightDirection), normalizedNormal));
		vec3 p_specular = pointLight[i].intensity * (1/attenuation) * material.k_Specular * pow(max(0.0f, dot(r2, normalize(view))), material.alpha);
		pLight += (texture(material.texture_diffuse0, TexCoord).rgb * p_diffuse + p_specular);
	}

	return pLight;
}


vec3 calcSpotLight(SpotLight spotLight[MAX_SpotLights], vec3 normalizedNormal, vec3 view) {
	vec3 sLight = vec3(0.0f, 0.0f, 0.0f);

	for(int i = 0; i < numberOfSLights; i++){
		vec3 sp = position_world - spotLight[i].position;
		vec3 r2 = normalize(reflect(sp, normalizedNormal));
		float angle = dot(normalize(sp), normalize(spotLight[i].direction));
		float d = distance(position_world, spotLight[i].position);
		float attenuation = spotLight[i].constant + d * spotLight[i].linear + pow(d, 2) * spotLight[i].quadratic;

		if(angle > spotLight[i].innerAngle) {
			vec3 s_diffuse = spotLight[i].intensity * (1/attenuation) * material.k_Diffuse * max(0.0f, dot(-normalize(sp), normalizedNormal));
			vec3 s_specular = spotLight[i].intensity * (1/attenuation) * material.k_Specular * pow(max(0.0f, dot(r2, normalize(view))), material.alpha);
			sLight += (texture(material.texture_diffuse0, TexCoord).rgb * s_diffuse + s_specular);
		}

		if(angle <= spotLight[i].innerAngle && angle >= spotLight[i].outerAngle) {
			vec3 s_diffuse = spotLight[i].intensity * (1/attenuation) * material.k_Diffuse * max(0.0f, dot(-normalize(sp), normalizedNormal));
			vec3 s_specular = spotLight[i].intensity * (1/attenuation) * material.k_Specular * pow(max(0.0f, dot(r2, normalize(view))), material.alpha);
			sLight += (texture(material.texture_diffuse0, TexCoord).rgb * s_diffuse + s_specular) * smoothstep(spotLight[i].outerAngle, spotLight[i].innerAngle, angle);
		}
	}

	return sLight;
}